package jsr.daemon;

import java.util.*;

/**
 * synchronized helper class for implementing an attribute mapping for daemons based on the HashMap class
 * providing for a chronological enumeration returning attributes in their original setting order
 * so that a command line input sequence is retained for subsequent attribute access
 * with runtime behaviour and null value support inherited from the HashMap implementation
 *
 * @author   Thomas Kopp
 * @version  0.9, 2000-11-16
 */
public class DaemonAttributes {

    /**
     * static inner helper class for binding attribute values and additional references
     */
    protected static class DaemonAttribute {
        
        /**
         * reference to the predecessor of this attribute
         */
        private DaemonAttribute previous = null;

        /**
         * reference to the successor of this attribute
         */
        private DaemonAttribute next = null;

        /**
         * key of this attribute
         */
        private String key = null;

        /**
         * value of this attribute
         */
        private Object value = null;

        
        /**
         * contructor of an attribute
         * @param key the key of this attribute
         * @param value the value of this attribute
         */
        protected DaemonAttribute(String key, Object value) {
            this.key = key;
            this.value = value;
        }
        
        /**
         * method for getting the key of this attribute
         * @return key of this attribute 
         */
        protected String getKey() {
            return key;
        }
        
        /**
         * method for getting the value of this attribute
         * @return value of this attribute 
         */
        protected Object getValue() {
            return value;
        }
        
        /**
         * method for setting the predecessor of this attribute
         * @param previous reference to the predecessor of this attribute 
         */
        protected void setPrevious(DaemonAttribute previous) {
            this.previous = previous;
        }
        
        /**
         * method for getting the predecessor of this attribute
         * @return reference to the predecessor of this attribute 
         */
        protected DaemonAttribute getPrevious() {
            return previous;
        }

        /**
         * method for setting the successor of this attribute
         * @param next reference to the successor of this attribute 
         */
        protected void setNext(DaemonAttribute next) {
            this.next = next;
        }
        
        /**
         * method for getting the successor of this attribute
         * @return reference to the successor of this attribute 
         */
        protected DaemonAttribute getNext() {
            return next;
        }
    }
    
    
    /**
     * mapping of the attributes stored
     */
    private Map environment = null;  
    
    /**
     * first attribute of the mapping according to a chronological order
     */
    private DaemonAttribute head = null;
   
    /**
     * last attribute of the mapping according to a chronological order
     */
    private DaemonAttribute tail = null;

   
    /** 
     * default constructor for this attribute mapping
     */
    public DaemonAttributes() {
        environment = new HashMap();
    }
   
    /** 
     * constructor for this attribute mapping with size estimation
     * @param initialCapacity estimated number of attributes in this mapping
     */
    public DaemonAttributes(int initialCapacity) {
        environment = new HashMap(initialCapacity);
    }
   
    /**
     * method for binding an attribute in this mapping
     * @param key the key of the attribute to bind
     * @param value the value of the attribute to bind
     */
    public synchronized void put(String key, Object value) {
        DaemonAttribute current = new DaemonAttribute(key, value);
        Object ancestor = environment.put(key, current);
        if (null != ancestor) {
           DaemonAttribute attribute = (DaemonAttribute)ancestor;
           DaemonAttribute previous = attribute.getPrevious();
           if (null != previous) {
              current.setPrevious(previous);
              previous.setNext(current);
           }
           else head = current;
           DaemonAttribute next = attribute.getNext();
           if (null != next) {
              current.setNext(next);
              next.setPrevious(current);
           }
           else tail = current;
        }
        else {
               if (null == head) head = current;
               if (null != tail) {
                  tail.setNext(current);
                  current.setPrevious(tail);
               }
               tail = current;
        }
    }

    /*
     * method for retrieving an attribute bound in this mapping
     * @param key the key of the attribute to retrieve
     * @return returns an attribute value bound under the specified key
     */
    public synchronized Object get(String key) {
        Object current = environment.get(key);
        if (current != null) return ((DaemonAttribute)current).getValue();
        else return null;
    }
   
    /**
     * method for removing an attribute bound in this mapping
     * @param key the key of the attribute to remove
     */
    public synchronized void remove(String key) {
        Object ancestor = environment.remove(key);
        if (null != ancestor) {
           DaemonAttribute attribute = (DaemonAttribute)ancestor;
           DaemonAttribute previous = attribute.getPrevious();
           DaemonAttribute next = attribute.getNext();
           if (null != next) next.setPrevious(previous);
           else tail = previous;
           if (null != previous) previous.setNext(next);
           else head = next;
        }
    }

    /*
     * method for retrieving the attribute keys of this mapping
     * access is synchronized but using the enumeration returned requires the whole access to be synchronized
     * @return enumeration of the keys of this mapping in their chronological setting order
     */
    public Enumeration keys() {
        return new Enumeration() {
            
            /**
             * reference to the current key of this enumeration
             */
            DaemonAttribute current = head;

            /**
             * indicator for more keys in this enumeration
             * @return true if and only if there are more keys in this enumeration
             */
            public boolean hasMoreElements() {
                return (null != current);
            }
            
            /**
             * method for accessing the next element in this enumeration
             * @return next key in this enumeration
             */
            public synchronized Object nextElement() {
                if (current != null) {
                   String key = current.getKey();
                   current = current.getNext();
                   return key;
                }
                else throw new NoSuchElementException();
            }
        };
    }
}
